const { Injectable } = require('flitter-di')
const { withTimeout } = require('./util')
const { Job: QueueJob, QueueEvents } = require('bullmq')

/**
 * Base class for Flitter job definitions.
 * @extends {module:flitter-di/src/Injectable~Injectable}
 */
class Job extends Injectable {
    static canonical_name = 'Job'
    static default_queue = undefined
    #queue_job = undefined

    /**
     * Get the services required by this unit.
     * Currently requires redis and jobs.
     * @returns {string[]}
     */
    static get services(){
        return [...super.services, 'redis', 'jobs', 'output']
    }

    static async enqueue(data, queue_name = undefined) {
        if ( !queue_name ) {
            queue_name = this.default_queue
        }

        if ( !queue_name ) {
            throw new Error('Unable to enqueue job. No queue or default queue specified.')
        }

        const queue = this.prototype.jobs.queue(queue_name)
        const queue_job = await queue.add(this.canonical_name, data)
        const job = new this()

        await job.fill_from_queue(queue_job)
        return job
    }

    async fill_from_queue(queue_job) {
        this.#queue_job = queue_job
        this.data = queue_job.data
        this.name = queue_job.name
        this.id = this.#queue_job.id
        this.queue = this.#queue_job.queue
    }

    /**
     * Called when the job is executed. The job is
     * completed when the promise of this function
     * resolves.
     * @param {bullmq/Job} job - the instance of the BullMQ job
     * @returns {Promise<void>}
     */
    async execute(job) {

    }

    toJSON() {
        return this.#queue_job.toJSON()
    }

    asJSON() {
        return this.#queue_job.asJSON()
    }

    async update(data) {
        await this.#queue_job.update(data)
    }

    async progress(progress) {
        await this.#queue_job.updateProgress(progress)
    }

    _log_prefix(type = 'log') {
        return `[${this.#queue_job.name} @ ${this.#queue_job.id}] [${type}] [${(new Date).toISOString()}] - `
    }

    info(string) {
        const prefix = this._log_prefix('info')
        this.output.info(`${prefix}${string}`)
        return this.#queue_job.log(`${prefix}${string}`)
    }

    success(string) {
        const prefix = this._log_prefix('success')
        this.output.success(`${prefix}${string}`)
        return this.#queue_job.log(`${prefix}${string}`)
    }

    error(string) {
        const prefix = this._log_prefix('error')
        this.output.error(`${prefix}${string}`)
        return this.#queue_job.log(`${prefix}${string}`)
    }

    debug(string) {
        const prefix = this._log_prefix('debug')
        this.output.debug(`${prefix}${string}`)
        return this.#queue_job.log(`${prefix}${string}`)
    }

    warn(string) {
        const prefix = this._log_prefix('warn')
        this.output.warn(`${prefix}${string}`)
        return this.#queue_job.log(`${prefix}${string}`)
    }

    async remove() {
        await this.#queue_job.remove()
    }

    async is_completed() {
        return this.#queue_job.isCompleted()
    }

    async is_failed() {
        return this.#queue_job.isFailed()
    }

    async is_delayed() {
        return this.#queue_job.isDelayed()
    }

    async is_active() {
        return this.#queue_job.isActive()
    }

    async isWaiting() {
        return this.#queue_job.isWaiting()
    }

    async state() {
        return this.#queue_job.getState()
    }

    async promote() {
        await this.#queue_job.promote()
    }

    async retry(state = 'failed') {
        await this.#queue_job.retry(state)
    }

    async sleep(ms) {
        await new Promise(res => {
            setTimeout(() => res(), ms)
        })
    }

    fail_error() {
        return new Error(this.#queue_job.failedReason)
    }

    return_value() {
        return this.#queue_job.returnvalue
    }

    toPromise() {
        return this.#queue_job.waitUntilFinished(new QueueEvents(this.queue.name))
    }

    toTimeout(ttl) {
        return withTimeout(ttl, this.toPromise())
    }
}

module.exports = exports = Job
