const { CanonicalUnit } = require('libflitter')
const { Queue } = require('bullmq')
const Job = require('./Job')
const WorkerDirective = require('./directive/WorkerDirective')
const ncp = require('ncp')
const path = require('path')
const express = require('express')

/**
 * Canonically registers job definitions and sets
 * up BullMQ queueing.
 * @extends {module:libflitter/canon/CanonicalUnit~CanonicalUnit}
 */
class JobsUnit extends CanonicalUnit {
    /**
     * Get the name of the service provided by this unit. ("jobs")
     * @returns {string}
     */
    static get name() { return 'jobs' }

    /**
     * Get the services required by this unit.
     * Currently configs, redis, output.
     * @returns {string[]}
     */
    static get services() {
        return [...super.services, 'configs', 'redis', 'output', 'utility', 'app']
    }

    /**
     * Mapping of queue name to BullMQ/Queue instance.
     * @private
     * @type {{}}
     */
    #queues = {}

    /**
     * Get a mapping of the CLI directives provided by this unit.
     * @returns {object}
     */
    directives() {
        return {
            WorkerDirective,
        }
    }

    /**
     * Initializes the unit.
     * @param [base_dir = './app/jobs']
     */
    constructor(base_dir = './app/jobs') {
        super(base_dir)

        /**
         * Name of one canonical item. ("job")
         * @type {string}
         */
        this.canonical_item = 'job'

        /**
         * File extension of canonical file definitions. (".job.js")
         * @type {string}
         */
        this.suffix = '.job.js'
    }

    /**
     * Get the jobs config.
     * @returns {*}
     */
    config() {
        return this.configs.get('jobs')
    }

    /**
     * Get a configured BullMQ/Queue instance by name.
     * @param {string} name - the name of a configured job queue
     * @returns {bullmq/Queue} - the queue instance
     */
    queue(name) {
        return this.#queues[name]
    }

    /**
     * Initializes the unit by loading canonical items and
     * connecting the job queue instances to Redis.
     * @param {module:libflitter/app/FlitterApp~FlitterApp} app
     * @returns {Promise<void>}
     */
    async go(app) {
        await super.go(app)
        const config = this.config()
        const redis_config = this.redis.config()

        if ( Array.isArray(config?.queues) ) {
            this.output.info('Connecting job queues...')
            for ( const queue_name of config.queues ) {
                this.output.debug(`Connecting to job queue ${queue_name}...`)
                this.#queues[queue_name] = new Queue(queue_name, {
                    connection: redis_config.server,
                })
            }
        }

        if ( config?.connector?.enabled ) {
            const api = this.get_connector_api()

            if ( config?.connector?.mount ) {
                this.app.express.use(config.connector.mount, api)
            }
        }
    }

    /**
     * Get the templates provided by this unit.
     */
    templates() {
        return {
            job: {
                template: require('../template/job'),
                directory: this.directory,
                extension: '.job.js',
            },
        }
    }

    /**
     * Instantiates one canonical item.
     * @param {module:libflitter/app/FlitterApp~FlitterApp} app
     * @param {string} name
     * @param {function} instance
     * @returns {Promise<function|void>}
     */
    async init_canonical_file({ app, name, instance }) {
        if ( !(instance.prototype instanceof Job) ) {
            this.output.warning(`Skipping malformed job ${name}. Jobs must extend flitter-jobs/src/Job.`)
            return
        }

        instance.canonical_name = name
        return instance
    }

    /**
     * Cleans up the resources provided by this unit by
     * disconnecting the job queue instances from Redis.
     * @param {module:libflitter/app/FlitterApp~FlitterApp} app
     * @returns {Promise<void>}
     */
    async cleanup(app) {
        this.output.info('Disconnecting job queues...')
        const queues = Object.values(this.#queues)
        await Promise.all(queues.map(x => x.close()))
    }

    /**
     * Deploy the configuration files for the jobs unit.
     * @returns {Promise<void>}
     */
    async deploy() {
        const deploy_dir = path.resolve(__dirname, '..', 'deploy')
        const base_dir = this.utility.path()

        this.output.info('Deploying jobs configurations...', 0)
        await new Promise((res, rej) => {
            ncp(deploy_dir, base_dir, err => {
                if ( err ) rej(err)
                else res()
            })
        })
    }

    get_connector_api() {
        const router = express.Router()

        const get_config = () => {
            const conf = this.configs.get('jobs.connector')
            return {
                enabled: false,
                ...(conf || {})
            }
        }

        const send_api = (res, status, message, data = {}) => {
            this.output.info(`Connector API Response (HTTP ${status}) - ${message}`)
            res.status(status)
            res.setHeader('Content-Type', 'application/json')

            return res.send({
                status,
                message,
                data,
            })
        }

        router.use((req, res, next) => {
            if ( !req.query || !req.query.accessSecret || !get_config().secret ) {
                return send_api(res, 401, 'Missing accessSecret query')
            }

            if ( req.query.accessSecret !== get_config().secret ) {
                return send_api(res, 401, 'Invalid accessSecret')
            }

            return next()
        })

        const get_queue = (req, res, next) => {
            if ( 'queue' in req.params ) {
                const queue = this.queue(req.params.queue)
                if ( !queue ) {
                    return send_api(res, 400, 'Invalid queue name')
                }

                req.queue = queue
                return queue
            }
        }

        router.get('/queues', (req, res, next) => {
            return send_api(res, 200, 'OK', (this.config()?.queues || []).map(queue_name => {
                return { queue_name }
            }))
        })

        router.get('/queues/:queue/count/backlog', async (req, res, next) => {
            if ( !get_queue(req, res, next) ) return;
            return send_api(res, 200, 'OK', await req.queue.count())
        })

        router.get('/queues/:queue/count/completed', async (req, res, next) => {
            if ( !get_queue(req, res, next) ) return;
            return send_api(res, 200, 'OK', await req.queue.getCompletedCount())
        })

        router.get('/queues/:queue/count/failed', async (req, res, next) => {
            if ( !get_queue(req, res, next) ) return;
            return send_api(res, 200, 'OK', await req.queue.getFailedCount())
        })

        router.get('/queues/:queue/count/delayed', async (req, res, next) => {
            if ( !get_queue(req, res, next) ) return;
            return send_api(res, 200, 'OK', await req.queue.getDelayedCount())
        })

        router.get('/queues/:queue/count/active', async (req, res, next) => {
            if ( !get_queue(req, res, next) ) return;
            return send_api(res, 200, 'OK', await req.queue.getActiveCount())
        })

        router.get('/queues/:queue/count/waiting', async (req, res, next) => {
            if ( !get_queue(req, res, next) ) return;
            return send_api(res, 200, 'OK', await req.queue.getWaitingCount())
        })

        const job_to_json = (job) => {
            return job.asJSON()
        }

        router.get('/queues/:queue/jobs', async (req, res, next) => {
            if ( !get_queue(req, res, next) ) return;
            const { start = 0, end = -1 } = req.query
            const jobs = await req.queue.getJobs(['completed', 'failed', 'delayed', 'repeat', 'active', 'wait', 'paused'], start, end)

            jobs.sort((a, b) => Number(b.id) - Number(a.id))

            return send_api(res, 200, 'OK', jobs.map(x => job_to_json(x)))
        })

        router.get('/queues/:queue/jobs/:jobType', async (req, res, next) => {
            if ( !get_queue(req, res, next) ) return;
            const { start = 0, end = -1 } = req.query
            const jobs = await req.queue.getJobs([req.params.jobType], start, end)

            jobs.sort((a, b) => Number(b.id) - Number(a.id))

            return send_api(res, 200, 'OK', jobs.map(x => job_to_json(x)))
        })

        router.get('/queues/:queue/workers', async (req, res, next) => {
            if ( !get_queue(req, res, next) ) return;
            return send_api(res, 200, 'OK', await req.queue.getWorkers())
        })

        router.get('/queues/:queue/job/:jobId', async (req, res, next) => {
            if ( !get_queue(req, res, next) ) return;
            const job = await req.queue.getJob(req.params.jobId)
            if ( !job ) {
                return send_api(res, 404, 'Job not found with that ID.')
            }

            return send_api(res, 200, 'OK', job_to_json(job))
        })

        router.get('/queues/:queue/job/:jobId/logs', async (req, res, next) => {
            if ( !get_queue(req, res, next) ) return;
            const { start = 0, end = -1 } = req.query
            return send_api(res, 200, 'OK', await req.queue.getJobLogs(req.params.jobId, start, end))
        })

        router.use((req, res, next) => {
            send_api(res, 404, 'Not Found')
        })

        return router
    }
}

module.exports = exports = JobsUnit
