const Directive = require('flitter-cli/Directive')
const PositionalOption = require('flitter-cli/options/PositionalOption')
const { Worker } = require('bullmq')

/**
 * flitter-cli directive to launch a worker process.
 * @extends {module:flitter-cli/Directive~Directive}
 */
class WorkerDirective extends Directive {
    /**
     * Get the services required by this unit.
     * Currently configs, jobs, and output.
     * @returns {string[]}
     */
    static get services() {
        return [...super.services, 'configs', 'jobs', 'output']
    }

    /**
     * Get the command name of this directive. ("worker")
     * @returns {string}
     */
    static name() {
        return 'worker'
    }

    /**
     * Get the help message for this directive.
     * @returns {string}
     */
    static help() {
        return "run this instance as a job queue worker: <worker name>"
    }

    /**
     * Build the options required by this directive.
     * Currently has one positional option, worker name.
     * @returns {Array<module:flitter-cli/options/Option~Option>}
     */
    static options() {
        const config = this.prototype.configs.get('jobs.workers')
        const values = Object.keys(config)
        const worker_option = new PositionalOption('worker name', 'worker configuration to process as')
        worker_option.whitelist(...values)
        return [worker_option]
    }

    /**
     * Get the jobs config.
     * @returns {*}
     */
    config() {
        return this.configs.get('jobs')
    }

    /**
     * Handle an invocation of this directive. Starts the
     * queue workers configured for this worker's name based
     * on the jobs.workers config.
     *
     * Press ^C to exit.
     *
     * @param {module:libflitter/app/FlitterApp~FlitterApp} app
     * @param {Array<string>} argv - the input arguments
     * @returns {Promise<void>}
     */
    async handle(app, argv) {
        const [ worker_name ] = argv
        const config = this.config()
        const redis_config = this.configs.get('redis')
        const worker_config = config.workers[worker_name]
        const queue_workers = {}

        for ( const queue_name of worker_config.queues ) {
            this.info(`Starting queue ${queue_name}...`)
            queue_workers[queue_name] = new Worker(queue_name, async job => {
                this.output.debug(`Processing new job: ${job.name}`)
                const JobClass = this.jobs.get(job.name)
                const job_inst = new JobClass()

                await job_inst.fill_from_queue(job)
                job_inst.info(`Started execution on worker ${worker_name} in queue ${queue_name}.`)

                try {
                    const ret = await job_inst.execute(job)
                    job_inst.success('Finished execution.')
                    return ret
                } catch(e) {
                    job_inst.error('Error encountered!')
                    job_inst.error(e)

                    throw e
                }
            }, { connection: redis_config.server })
        }

        await new Promise(res => {
            process.on('SIGINT', async () => {
                this.info('Stopping queues...')
                await Promise.all(Object.values(queue_workers).map(x => x.close()))
                res()
            })
        })
    }
}

module.exports = exports = WorkerDirective
